/*	
HEADER:
TITLE:		Fast Fourier Transform;
DATE:		05/18/1985;
DESCRIPTION:	"Performs fast fourier transform using method described
		by E. O. Brigham.  For details of the method, refer
		to Brigham's book. THE FAST FOURIER TRANSFORM";	
KEYWORDS: 	Fourier, transform;
FILENAME:	INT_FFT.C;
WARNINGS:	
  "This program is self-contained, all that is needed is a manner of getting
  the data into the input array.
  The  transformed data will reside in these two arrays upon return with the 
  original data being destroyed."
AUTHORS:	Jim Pisano, Michael Schippling;
COMPILERS:	Aztec.
REFERENCES:	AUTHORS:	E. O. Brigham;
		TITLE:		"THE FAST FOURIER TRANSFORM";
		CITATION:	"";
	ENDREF
*/

/*	file name fft.c
*	program name fft() ... Fast Fourier Transform
*
*	Perform fast fourier transform using method described by E. O. Brigham.
*  For details of the method, refer to Brigham's book 
*
*	Translated to C from FORTRAN by
*
*		Jim Pisano
*		P.O. Box 3134
*		University Station
*		Charlottesville, VA 22903
*
*	Converted to any data type and debugged by
*
*		Michael Schippling
*		GS IMPROVISION
*		1690 65th St.
*		Emeryville, CA  94608
*
*  This program is in the public domain & may be used by anyone for commercial
*  or non-commercial purposes.
*
*  This program is self-contained, all that is needed is a manner of getting
*   the data into the input array.
*  The transformed data will reside in this array upon return with the 
*   original data being destroyed.
*/

#include	<stdio.h>
#include	<math.h>	    /* declare math functions to use */
/* #include	<macros.h>	    /* common macros */
#include	"fft.h"		    /* data types */

/* #define TBLDEBUG	/* print sin/cos table during generation */

static unsigned N_pts, Nu;	/* number of points to analyze, and log2 n */
static FFT_DATA *Atable;	/* Beginning of cos/sin argument table */

/* Simple exchange routine where COMPLEX array elements x and y are swapped.
 *  Depends on the compiler's ability to assign structures.
 *  If this is not supported, explicitly assign real and imaginary parts. */
#define swap(r,x,y)	{COMPLEX t; t=r[x]; r[x]=r[y]; r[y]=t;}

/* get elements from sin/cos table */
#define getcos(j)	( (FC) *(Atable+(j)) )
#define getsin(j)	( (FC) *(Atable+(j)+1) )


static unsigned
bit_swap (i)
	/* Bit swaping routine in which the bit pattern
	 *  of the integer i is reordered.  Uses global Nu.
	 * Ret:	i in bit reversed order. */
    unsigned i;
{
	unsigned up, dn, cnt, ib;

	up = 1;
	dn = 1 << (Nu-1);
	for (ib = 0, cnt = Nu; cnt; --cnt)
	{
		if (i & up)	ib |= dn;
		up <<= 1;
		dn >>= 1;
	}

	return (ib);
}


FFT_DATA *
sin_arg_table (n_pts, nu)
	/* Alloc and build an argument table of cos and sin values 
	 *  for even "j" values from 0 to n_pts-1.
	 * The table entries are converted from floating point to the
	 *  data type used by the FFT function.
	 * Table has alternating cos-sin values in even-odd locations.
	 * Sets these globals globals:
	 *	FFT_DATA *Atable;	points to beginning of table.
	 *	unsigned N_pts;		number of data points.
	 *	unsigned Nu;		log2 of N_pts.
	 * Use define FIXMULT to convert float to fixed values if necessary.
	 * Ret:	pointer to table or NULL if failed.
	 *	 (pointer should be freed after use). */
    unsigned n_pts;	/* number of points in sample */
    unsigned nu;	/* log 2 of n_pts */
{
	unsigned j, ib, np1;	/* temp values */
	FFT_DATA *atp;		/* table to fill (n_pts shorts in length) */
	double arg;		/* argument for sin and cos */
	FFT_DATA *malloc();

	/* allocate memory for table */
	if ( (Atable = atp = malloc (n_pts * sizeof(FFT_DATA))) ) 
	{
		N_pts = n_pts;
		Nu = nu;

		for ( j = 0, np1 = n_pts-1; j < np1; j += 2 )
		{
			ib = bit_swap (j);
			arg = 2 * PI * ib / n_pts;
			*atp++ = (FFT_DATA) (FIXMULT * cos (arg));
			*atp++ = (FFT_DATA) (FIXMULT * sin (arg));
#ifdef TBLDEBUG
#ifdef FLOATDATA
	fprintf (stderr, "%d %f  cos %f,\tsin %f\n",
#endif
#ifdef LONGDATA
	fprintf (stderr, "%d %f  cos %ld,\tsin %ld\n",
#endif
#ifdef SHORTDATA
	fprintf (stderr, "%d %f  cos %d,\tsin %d\n",
#endif
		j, arg, *(atp-2), *(atp-1) );
#endif
		}
	}
	return (Atable);
}


fft (data, inv)
    	/* Compute the forward FFT of the COMPLEX data points for N_pts.
	 * Returns results in data array and does computations there.
	 * The extern FFT_DATA Atable[] sin/cos argument array,
	 *  and the extern N_pts and Nu number of points and log2 N_pts,
	 *  should be already filled in using sin_arg_table().
	 * (See fft.h header file for details on typedefs). */
    COMPLEX *data;	/* input data and FFT result array */
    int inv;		/* 1 = forward, -1 = reverse transform */
{
	unsigned j, l, i, ib, k, nu1, nu12, n2, np1;	/* temp values */
	COMPLEX *rp1, *rp2;	/* pointers to elements in calculations */
	FC tr, ti;		/* intermediate values in calcs. */
	FC c, s;		/* cosine & sine components of Fourier trans. */

	nu1 = Nu;
	n2 = N_pts;
	np1 = N_pts - 1;

	/* Calculate the components of the Fourier series of the function */
	for( l = 0; l != Nu; l++ )
	{
		k = 0;
		nu1--;
		n2 >>= 1;
		do
		{
			nu12 = 1 << nu1;
			for( i = 0; i != n2; i++, k++ )
			{	
				j = k / nu12;
				c = getcos (j);
				s = inv * getsin (j);
				rp1 = data + k;
				rp2 = rp1 + n2;
				tr = ( (rp2->real * c) +
				       (rp2->imag * s) ) / FIXMULT;
				ti = ( (rp2->imag * c) -
				       (rp2->real * s) ) / FIXMULT;
				rp2->real = rp1->real - tr;
				rp2->imag = rp1->imag - ti;
				rp1->real = rp1->real + tr;
				rp1->imag = rp1->imag + ti;
			}
			k +=  n2;
		} while( k < np1);
	}	

	for( k = 0; k != N_pts; k++ )
	{
		ib = bit_swap (k);
		if (ib > k)
		{
			swap (data, k, ib);
		}
	}

	/* If calculating the inverse transform,
	 *  must divide the data by the number of data points. */
	if( inv == REV_FFT)
	{
		for( rp1 = data, k = 0; k != N_pts; rp1++, k++)
		{
			rp1->real /= N_pts;
			rp1->imag /= N_pts;
		} 
	}

	return (0);
}
